from typing import Sequence
from collections import OrderedDict
import logging

import numpy as np
import xarray as xr
from tqdm import tqdm

tqdm.monitor_interval = 0
log = logging.getLogger()


class CoordinateIterator:
    def __init__(self, coord_labels: Sequence[str], ds: xr.Dataset, desc='', **progress):
        """
        An flat iterator across multiple coordinates of an xr.Dataset.

        Parameters
        ----------
        coord_labels: list of str's
            The names (in order) of the coordinates to iterate across in a flat fashion.
        ds: xr.Dataset
            The xr.Dataset whose coordinates we're going to iterate across.
        desc: str
            A description that will be displayed infront of the tqdm progress bar.
        **progress:
            Keyword arguments mapping coordinate names to the index to start at.
        """
        self.coords = [ds[coord] for coord in coord_labels]
        self.indexes = OrderedDict([(coord.name, 0) for coord in self.coords])
        # assert progress keys are a subset of indexes keys
        assert set(progress.keys()).issubset(set(self.indexes.keys()))
        self.indexes.update(progress)
        for k, v in self.indexes.items():
            if v > 0:
                log.info(f"Continuing coordinate '{k}' at {v}")
        self.dims = [len(ds[coord]) for coord in coord_labels]
        self.initial = True
        self.pbar = tqdm(total=len(self), desc=desc)
        self.pbar.update(self.flat_index)

    @property
    def flat_index(self):
        last_dimsize = 1
        idx = 0
        for d, coord in enumerate(self.coords):
            idx += last_dimsize * self.indexes[coord.name]
            last_dimsize *= self.dims[d]
        return idx

    def __len__(self):
        if len(self.dims) > 0:
            return np.product(self.dims)
        else:
            return 1

    def __iter__(self):
        return self

    def update_postfix(self):
        postfix = 'Indexes: '
        for d, coord in enumerate(self.coords):
            postfix += f"{coord.name}={self.indexes[coord.name]} "
        self.pbar.set_postfix_str(postfix.strip())

    def __next__(self):
        if self.initial:
            self.initial = False
            self.update_postfix()
            return {n: slice(v, v + 1) for n, v in self.indexes.items()}
        self.pbar.update()
        carry = 1
        if len(self.coords) > 0:
            for d, coord in enumerate(self.coords):
                self.indexes[coord.name] += carry
                if self.indexes[coord.name] >= self.dims[d]:
                    self.indexes[coord.name] = 0
                    carry = 1
                    if d == len(self.coords) - 1:
                        raise StopIteration
                else:
                    carry = 0
        else:
            raise StopIteration

        self.update_postfix()
        return {n: slice(v, v + 1) for n, v in self.indexes.items()}


class ExpectedCoordIterator(CoordinateIterator):
    def __init__(self, table: xr.Dataset, **progress):
        """
        A subclass of CoordinateIterator that is expecting certain coordinate names (or 'standard_name' or 'long_name'
        attributes). If an instance of this class detects any self.expected_coordinates it will iterate across those
        dimensions.
        """
        # build arguments for CoordinateIterator
        self.coord_labels = []
        for expected in self.expected_coords:
            try:
                da = table.cf[expected]
                if len(da.dims) > 0:
                    self.coord_labels.append(da.name)
            except KeyError:
                pass
        appropriate_progress = {}
        for coord in self.coord_labels:
            if coord in progress:
                appropriate_progress[coord] = progress[coord]
        super().__init__(self.coord_labels, table, self.desc, **appropriate_progress)


class ClearAtmosphereVariables(ExpectedCoordIterator):
    desc = 'Clear-sky atmosphere variables'
    expected_coords = ['surface_albedo', 'surface_air_pressure']


class CloudyAtmosphereVariables(ExpectedCoordIterator):
    desc = 'Cloudy-sky atmosphere variables'
    expected_coords = ['cloud_albedo', 'air_pressure_at_cloud_top', 'surface_air_pressure']


class GeometricVariables(ExpectedCoordIterator):
    desc = 'Geometric variables'
    expected_coords = ['solar_zenith_angle']


if __name__ == '__main__':
    import unittest

    ds = xr.Dataset()
    ds.coords['x'] = np.array([0, 1, 2, 3, 4])
    ds.coords['y'] = np.array([0, 1, 2])
    ds['data'] = xr.DataArray(np.array([  # (x+1)*(y+1) == data
        [1, 2, 3, 4, 5],    # 0 x1
        [2, 4, 6, 8, 10],   # 1 x2
        [3, 6, 9, 12, 15],  # 2 x3
    ]), coords=[ds.coords['y'], ds.coords['x']])

    class TestCoordinateIterator(unittest.TestCase):
        def test_ordered(self):
            y = -1
            x = 0
            min_value = 100
            max_value = 0
            for yx in CoordinateIterator(['y', 'x'], ds, 'Test iteration 1'):
                y += 1
                if y == 3:
                    y = 0
                    x += 1
                self.assertEqual(yx['y'], y)
                self.assertEqual(yx['x'], x)
                self.assertEqual(ds['data'].isel(**yx).item(), (x + 1) * (y + 1))
                if ds['data'].isel(**yx).item() < min_value:
                    min_value = ds['data'].isel(**yx).item()
                if ds['data'].isel(**yx).item() > max_value:
                    max_value = ds['data'].isel(**yx).item()
            self.assertEqual(min_value, 1)
            self.assertEqual(max_value, 15)

        def test_out_of_order(self):
            y = 0
            x = -1
            min_value = 100
            max_value = 0
            for xy in CoordinateIterator(['x', 'y'], ds, 'Test 2 iteration'):
                x += 1
                if x == 5:
                    x = 0
                    y += 1
                self.assertEqual(xy['y'], y)
                self.assertEqual(xy['x'], x)
                self.assertEqual(ds['data'].isel(**xy).item(), (x + 1) * (y + 1))
                if ds['data'].isel(**xy).item() < min_value:
                    min_value = ds['data'].isel(**xy).item()
                if ds['data'].isel(**xy).item() > max_value:
                    max_value = ds['data'].isel(**xy).item()
            self.assertEqual(min_value, 1)
            self.assertEqual(max_value, 15)

        def test_with_progress(self):
            y = 1
            x = -1 + 2
            min_value = 100
            max_value = 0
            for xy in CoordinateIterator(['x', 'y'], ds, 'Test 2 iteration', x=2, y=1):
                x += 1
                if x == 5:
                    x = 0
                    y += 1
                self.assertEqual(xy['y'], y)
                self.assertEqual(xy['x'], x)
                self.assertEqual(ds['data'].isel(**xy).item(), (x + 1) * (y + 1))
                if ds['data'].isel(**xy).item() < min_value:
                    min_value = ds['data'].isel(**xy).item()
                if ds['data'].isel(**xy).item() > max_value:
                    max_value = ds['data'].isel(**xy).item()
            self.assertEqual(min_value, 3)
            self.assertEqual(max_value, 15)

    unittest.main()