import numpy as np
import xarray as xr
import argparse

from skamf.utility import get_response, parse_option_names


def add_option(table, external_name, internal_name, engine, value, overwrite=False, verbose=False):
    engine = engine.lower()
    engine = 'do' if engine == 'disco' else engine
    if engine not in ['hr', 'do', 'mc']:
        if verbose:
            print(f'unkown engine \'{engine}\'')
        return

    true_external_name = f'o_{external_name}'
    if not overwrite and true_external_name in table:
        if verbose:
            print(f'option \'{external_name}\' is already present in the table')
        return

    if not overwrite:
        for var in table.variables:
            options = parse_option_names(table, var)
            if engine in options and options[engine] == internal_name:
                if verbose:
                    print(f'engine option \'{internal_name}\' is already set in dataarray \'{var}\'')
                return

    value = np.atleast_1d(value)
    if value.size == 1:
        table[true_external_name] = value.item()
    else:
        n = len(value.shape)
        if n == 1:
            table[true_external_name] = (f'o_{external_name}_dim', value)
        else:
            dims = [f'o_{external_name}_{i+1}' for i in range(n)]
            table[true_external_name] = (dims, value)
    table[true_external_name].attrs['sasktran_option'] = f'{engine}: {internal_name}'

    if verbose:
        print(table[true_external_name])


if __name__ == '__main__':

    parser = argparse.ArgumentParser(description='Add an option to an amf table')
    parser.add_argument('filename', type=str, help='path to the location of the table.')
    parser.add_argument('-r', '--remove', type=str, help='enter the table name of the option to remove')
    parser.add_argument('-p', '--preset', nargs='+', help='use a preset helper')

    args = vars(parser.parse_args())

    if args['remove'] is not None:
        new_table = None
        with xr.open_dataset(args['filename']) as table:
            if f"o_{args['remove']}" in table:
                if 'sasktran_option' in table[f"o_{args['remove']}"].attrs:
                    new_table = xr.Dataset()
                    # copy over every data_var and coord that is not being removed - this is the only way to remove
                    # a dimension associated only with the removed option
                    for var, array in table.data_vars.items():
                        if var != f"o_{args['remove']}":
                            new_table[var] = (array.dims, array.values, array.attrs)
                    for var, array in table.coords.items():
                        if var != f"o_{args['remove']}":
                            new_table.coords[var] = (array.dims, array.values, array.attrs)
                    new_table.attrs = table.attrs
                    print(new_table)
                else:
                    print(f"'o_{args['remove']}' is not a sasktran option")
            else:
                print(f"'o_{args['remove']}' not found in table '{args['filename']}")
        if new_table is not None:
            new_table.to_netcdf(args['filename'], mode='w')

    elif args['preset'] is None:
        with xr.open_dataset(args['filename']) as table:
            table_name = get_response(
                varname='Option name (table)',
                prompt='Enter the name of the option as it will appear in the AMF table',
                default=None, mandatory=True, type=str
            )[0]
            engine = get_response(
                varname='Sasktran engine',
                prompt='Enter the name of the Sasktran engine this option is for',
                default=None, mandatory=True, type=str
            )[0]
            sasktran_name = get_response(
                varname='Option name (sasktran)',
                prompt='Enter the name of the option that is recognized by sasktran',
                default=None, mandatory=True, type=str
            )[0]
            value = get_response(
                varname='Option value',
                prompt='Enter the value(s) of the option',
                default=None, mandatory=True, type=float
            )

            add_option(table, table_name, sasktran_name, engine, value, verbose=True)
        table.to_netcdf(args['filename'], mode='a')
    else:
        with xr.open_dataset(args['filename']) as table:
            if isinstance(args['preset'], str):
                print('every preset needs at least one additional argument')
                exit()

            if args['preset'][0] == 'diffuse_heights':
                res1 = float(args['preset'][1])
                res2 = float(args['preset'][2] if len(args['preset']) > 2 else 1e3)
                heights = np.array([1e-6, *np.arange(res1 / 2.0, 12e3, res1), *np.arange(12e3 + res2 / 2.0, 1e5, res2)])
                add_option(table, 'diffuse_heights', 'manualdiffuseheights', 'hr', heights, verbose=True)
            elif args['preset'][0] == 'diffuse_profiles':
                num = int(args['preset'][1])
                add_option(table, 'diffuse_profiles', 'numdiffuseprofilesinplane', 'hr', num, verbose=True)
            elif args['preset'][0] == 'diffuse_incoming':
                res = [int(n) for n in args['preset'][1:5]]
                add_option(table, 'diffuse_incoming', 'diffuseincomingresolution', 'hr', res, verbose=True)
            elif args['preset'][0] == 'diffuse_outgoing':
                num = int(np.round(np.sqrt(float(args['preset'][1])))**2)
                add_option(table, 'diffuse_outgoing', 'numdiffuseoutgoing', 'hr', num, verbose=True)
            elif args['preset'][0] == 'num_streams':
                num = int(args['preset'][1])
                add_option(table, 'num_streams', 'numstreams', 'do', num, verbose=True)
            elif args['preset'][0] == 'num_layers':
                num = int(args['preset'][1])
                add_option(table, 'num_layers', 'numlayers', 'do', num, verbose=True)
            elif args['preset'][0] == 'photons':
                num = float(args['preset'][1])
                add_option(table, 'num_photons', 'setnumphotonsperlos', 'mc', num, verbose=True)
            elif args['preset'][0] == 'target':
                num = float(args['preset'][1])
                add_option(table, 'target_accuracy', 'settargetstd', 'mc', num, verbose=True)
            else:
                print(f"unknown preset {args['preset']}")
                exit()
        table.to_netcdf(args['filename'], mode='a')
